# -*- coding: utf-8 -*-
"""Basic auth for Nessie client."""
import requests


def login(base_url: str, username: str, password: str, timeout: int = 10, verify: bool = True) -> str:
    """Log into Nessie using basic auth.

    :param base_url: Nessie url
    :param username: username
    :param password: password
    :param timeout: optional timeout
    :param verify: If false ignore ssl errors
    :return: auth token
    """
    url = base_url + "/login"

    r = requests.post(url, data={"username": username, "password": password, "grant_type": "password"}, timeout=timeout, verify=verify)
    r.raise_for_status()
    return r.json()["token"]



# -*- coding: utf-8 -*-
"""Console script for nessie_client."""
import datetime
import sys
from collections import defaultdict
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple

import attr
import click
import confuse
from click import Option
from click import UsageError
from dateutil.tz import tzlocal

from . import __version__
from ._ref import handle_branch_tag
from .conf import build_config
from .conf import process
from .error import error_handler
from .error import NessieNotFoundException
from .expression_util import build_query_expression_for_commit_log_flags
from .expression_util import build_query_expression_for_contents_listing_flags
from .model import CommitMeta
from .model import CommitMetaSchema
from .model import Contents
from .model import ContentsSchema
from .model import Delete
from .model import Entries
from .model import Entry
from .model import EntrySchema
from .model import Operation
from .model import Put
from .nessie_client import _contents_key
from .nessie_client import _format_key
from .nessie_client import NessieClient


@attr.s(auto_attribs=True)
class ContextObject(object):
    """Click context object."""

    nessie: NessieClient
    verbose: bool
    json: bool


pass_client = click.make_pass_decorator(ContextObject)


def _print_version(ctx: Any, param: Any, value: Any) -> None:
    if not value or ctx.resilient_parsing:
        return
    click.echo("nessie version " + __version__)
    ctx.exit()


class MutuallyExclusiveOption(Option):
    """Only allow one option in a list to be set at once."""

    def __init__(self: "MutuallyExclusiveOption", *args: List, **kwargs: Dict) -> None:
        """Instantiated a mutually exclusive option."""
        self.mutually_exclusive = set(kwargs.pop("mutually_exclusive", []))
        super(MutuallyExclusiveOption, self).__init__(*args, **kwargs)  # type: ignore

    def handle_parse_result(self: "MutuallyExclusiveOption", ctx: click.Context, opts: Mapping, args: List) -> Tuple[Any, List[str]]:
        """Ensure mutually exclusive options are not used together."""
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise UsageError(
                "Illegal usage: `{}` is mutually exclusive with " "arguments `{}`.".format(self.name, ", ".join(self.mutually_exclusive))
            )

        return super(MutuallyExclusiveOption, self).handle_parse_result(ctx, opts, args)


class DefaultHelp(click.Command):
    """If no options are presented show help."""

    def __init__(self: "DefaultHelp", *args: List, **kwargs: Dict) -> None:
        """Ensure that help is shown if nothing else is selected."""
        context_settings = kwargs.setdefault("context_settings", {})
        if "help_option_names" not in context_settings:
            context_settings["help_option_names"] = ["-h", "--help"]
        self.help_flag = context_settings["help_option_names"][0]
        super(DefaultHelp, self).__init__(*args, **kwargs)  # type: ignore

    def parse_args(self: "DefaultHelp", ctx: click.Context, args: List) -> List:
        """Ensure that help is shown if nothing else is selected."""
        if not args:
            args = [self.help_flag]
        return super(DefaultHelp, self).parse_args(ctx, args)


@click.group()
@click.option("--json", is_flag=True, help="write output in json format.")
@click.option("-v", "--verbose", is_flag=True, help="Verbose output.")
@click.option("--endpoint", help="Optional endpoint, if different from config file.")
@click.option("--repo", help="Optional repository owner + name, if different from config file, separated by a slash.")
@click.option("--version", is_flag=True, callback=_print_version, expose_value=False, is_eager=True)
@click.pass_context
def cli(ctx: click.core.Context, json: bool, verbose: bool, endpoint: str, repo: str) -> None:
    """Nessie cli tool.

    Interact with Nessie branches and tables via the command line
    """
    try:
        cfg_map = {}
        if endpoint:
            cfg_map["endpoint"] = endpoint
        if repo:
            cfg_map["repo"] = repo
        config = build_config(cfg_map)
        nessie = NessieClient(config)
        ctx.obj = ContextObject(nessie, verbose, json)
    except confuse.exceptions.ConfigTypeError as e:
        raise click.ClickException(str(e))


@cli.group()
@pass_client
def remote(ctx: ContextObject) -> None:
    """Set and view remote endpoint."""
    pass


@cli.command("config", cls=DefaultHelp)
@click.option("--get", cls=MutuallyExclusiveOption, help="get config parameter", mutually_exclusive=["set", "list", "unset"])
@click.option("--add", cls=MutuallyExclusiveOption, help="set config parameter", mutually_exclusive=["get", "list", "unset"])
@click.option(
    "-l",
    "--list",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="list config parameters",
    mutually_exclusive=["set", "get", "unset"],
)
@click.option("--unset", cls=MutuallyExclusiveOption, help="unset config parameter", mutually_exclusive=["get", "list", "set"])
@click.option("--type", help="type to interpret config value to set or get. Allowed options: bool, int")
@click.argument("key", nargs=1, required=False)
@pass_client
@error_handler
def config(ctx: ContextObject, get: str, add: str, list: bool, unset: str, type: str, key: str) -> None:
    """Set and view config."""
    res = process(get, add, list, unset, key, type)
    click.echo(res)


@remote.command("show")
@pass_client
@error_handler
def show(ctx: ContextObject) -> None:
    """Show current remote."""
    click.echo("Remote URL: " + ctx.nessie._base_url)
    click.echo("Default branch: " + ctx.nessie.get_reference(None).name)
    click.echo("Remote branches: ")
    for i in ctx.nessie.list_references():
        click.echo("\t" + i.name)


@remote.command(name="add")
@click.argument("endpoint", nargs=1, required=True)
@pass_client
@error_handler
def set_(ctx: ContextObject, endpoint: str) -> None:
    """Set current remote."""
    click.echo(process(None, "endpoint", False, None, endpoint))


@remote.command("set-head")
@click.argument("head", nargs=1, required=True)
@click.option("-d", "--delete", is_flag=True, help="delete the default branch")
@pass_client
@error_handler
def set_head(ctx: ContextObject, head: str, delete: bool) -> None:
    """Set current default branch. If -d is passed it will remove the default branch."""
    if delete:
        click.echo(process(None, None, False, "default_branch", None))
    else:
        click.echo(process(None, "default_branch", False, None, head))


@cli.command("log")
@click.option("-r", "--ref", help="branch to list from. If not supplied the default branch from config is used")
@click.option("-n", "--number", help="number of log entries to return", type=int)
@click.option("--since", "--after", help="Only include commits newer than specific date")
@click.option("--until", "--before", help="Only include commits older than specific date")
@click.option(
    "--author",
    multiple=True,
    help="Limit commits to a specific author (this is the original committer). Supports specifying multiple authors to filter by.",
)
@click.option(
    "--committer",
    multiple=True,
    help="Limit commits to a specific committer (this is the logged in user/account who performed the commit). "
    "Supports specifying multiple committers to filter by.",
)
@click.argument("revision_range", nargs=1, required=False)
@click.argument("paths", nargs=-1, type=click.Path(exists=False), required=False)
@click.option(
    "--query",
    "--query-expression",
    "query_expression",
    multiple=False,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["author", "committer", "since", "until"],
    help="Allows advanced filtering using the Common Expression Language (CEL). "
    "An intro to CEL can be found at https://github.com/google/cel-spec/blob/master/doc/intro.md.\n"
    "Some examples with usable variables 'commit.author' (string) / 'commit.committer' (string) / 'commit.commitTime' (timestamp) / "
    "'commit.hash' (string) / 'commit.message' (string) / 'commit.properties' (map) are:\n"
    "commit.author=='nessie_author'\n"
    "commit.committer=='nessie_committer'\n"
    "timestamp(commit.commitTime) > timestamp('2021-06-21T10:39:17.977922Z')\n",
)
@pass_client
@error_handler
def log(  # noqa: C901
    ctx: ContextObject,
    ref: str,
    number: int,
    since: str,
    until: str,
    author: List[str],
    committer: List[str],
    revision_range: str,
    paths: Tuple[click.Path],
    query_expression: str,
) -> None:
    """Show commit log.

    REVISION_RANGE optional hash to start viewing log from. If of the form <start_hash>..<end_hash> only show log
    for given range on the particular ref that was provided

    PATHS optional list of paths. If given, only show commits which affected the given paths
    """
    if not ref:
        ref = ctx.nessie.get_default_branch()
    start_hash = None
    end_hash = None
    if revision_range:
        if ".." in revision_range:
            start_hash, end_hash = revision_range.split("..")
        else:
            end_hash = revision_range

    filtering_args: Any = {}
    if number:
        filtering_args["max"] = str(number)
    if start_hash:
        filtering_args["startHash"] = start_hash
    if end_hash:
        filtering_args["endHash"] = end_hash
    expr = build_query_expression_for_commit_log_flags(query_expression, author, committer, since, until)
    if expr:
        filtering_args["query_expression"] = expr

    # TODO: limiting by path is not yet supported.
    log_result = ctx.nessie.get_log(start_ref=ref, **filtering_args)
    if ctx.json:
        click.echo(CommitMetaSchema().dumps(log_result, many=True))
    else:
        click.echo_via_pager(_format_log_result(x) for x in log_result)


def _format_log_result(x: CommitMeta) -> str:
    result = click.style("commit {}\n".format(x.hash_), fg="yellow")
    result += click.style("Author: {} <{}>\n".format(x.committer, x.email))
    result += click.style("Date: {}\n".format(_format_time(x.commitTime)))
    result += click.style("\n\t{}\n\n".format(x.message))
    return result


def _format_time(dt: datetime.datetime) -> str:
    return dt.astimezone(tzlocal()).strftime("%c %z")


@cli.command(name="branch")
@click.option("-l", "--list", cls=MutuallyExclusiveOption, is_flag=True, help="list branches", mutually_exclusive=["delete"])
@click.option("-d", "--delete", cls=MutuallyExclusiveOption, is_flag=True, help="delete a branch", mutually_exclusive=["list"])
@click.option("-f", "--force", is_flag=True, help="force branch assignment")
@click.option("-c", "--condition", help="Conditional Hash. Only perform the action if branch currently points to condition.")
@click.argument("branch", nargs=1, required=False)
@click.argument("new_branch", nargs=1, required=False)
@pass_client
@error_handler
def branch_(
    ctx: ContextObject,
    list: bool,
    force: bool,
    delete: bool,
    branch: str,
    new_branch: str,
    condition: str,
) -> None:
    """Branch operations.

    BRANCH name of branch to list or create/assign

    NEW_BRANCH name of branch to assign from or rename to

    Examples:

        nessie branch -l -> list all branches

        nessie branch -l main -> list only main

        nessie branch -d main -> delete main

        nessie branch -> list all branches

        nessie branch main -> create branch main at current head

        nessie branch main test -> create branch main at head of test

        nessie branch -f main test -> assign main to head of test

    """
    results = handle_branch_tag(ctx.nessie, list, delete, branch, new_branch, True, ctx.json, force, ctx.verbose, condition)
    if ctx.json:
        click.echo(results)
    elif results:
        click.echo(results)
    else:
        click.echo()


@cli.command("tag")
@click.option("-l", "--list", cls=MutuallyExclusiveOption, is_flag=True, help="list branches", mutually_exclusive=["delete"])
@click.option("-d", "--delete", cls=MutuallyExclusiveOption, is_flag=True, help="delete a branches", mutually_exclusive=["list"])
@click.option("-f", "--force", is_flag=True, help="force branch assignment")
@click.option("-c", "--condition", help="Conditional Hash. Only perform the action if branch currently points to condition.")
@click.argument("tag_name", nargs=1, required=False)
@click.argument("new_tag", nargs=1, required=False)
@pass_client
@error_handler
def tag(ctx: ContextObject, list: bool, force: bool, delete: bool, tag_name: str, new_tag: str, condition: str) -> None:
    """Tag operations.

    TAG_NAME name of branch to list or create/assign

    NEW_TAG name of branch to assign from or rename to

    Examples:

        nessie tag -l -> list all tags

        nessie tag -l main -> list only main

        nessie tag -d main -> delete main

        nessie tag -> list all tags

        nessie tag main -> create tag xxx at current head

        nessie tag main test -> create tag xxx at head of test

        nessie tag -f main test -> assign xxx to head of test

    """
    results = handle_branch_tag(ctx.nessie, list, delete, tag_name, new_tag, False, ctx.json, force, ctx.verbose, condition)
    if ctx.json:
        click.echo(results)
    elif results:
        click.echo(results)
    else:
        click.echo()


@cli.command("merge")
@click.option("-b", "--branch", "onto_branch", help="branch to merge onto. If not supplied the default branch from config is used")
@click.argument("from_branch", nargs=1, required=False)
@click.option(
    "-f",
    "--force",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    mutually_exclusive=["condition"],
    help="force branch assignment",
)
@click.option(
    "-c",
    "--condition",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["force"],
    help="Conditional Hash. Only perform the action if branch currently points to condition.",
)
@pass_client
@error_handler
def merge(ctx: ContextObject, onto_branch: str, force: bool, condition: str, from_branch: str) -> None:
    """Merge FROM_BRANCH into current branch. FROM_BRANCH can be a hash or branch."""
    if not force and not condition:
        raise UsageError(
            """Either condition or force must be set. Condition should be set to a valid hash for concurrency
            control or force to ignore current state of Nessie Store."""
        )
    ctx.nessie.merge(from_branch, onto_branch if onto_branch else ctx.nessie.get_default_branch(), condition)
    click.echo()


@cli.command("cherry-pick")
@click.option("-b", "--branch", help="branch to cherry-pick onto. If not supplied the default branch from config is used")
@click.option(
    "-f",
    "--force",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    mutually_exclusive=["condition"],
    help="force branch assignment",
)
@click.option(
    "-c",
    "--condition",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["force"],
    help="Conditional Hash. Only perform the action if branch currently points to condition.",
)
@click.argument("hashes", nargs=-1, required=False)
@pass_client
@error_handler
def cherry_pick(ctx: ContextObject, branch: str, force: bool, condition: str, hashes: Tuple[str]) -> None:
    """Transplant HASHES onto current branch."""
    if not force and not condition:
        raise UsageError(
            """Either condition or force must be set. Condition should be set to a valid hash for concurrency
            control or force to ignore current state of Nessie Store."""
        )
    ctx.nessie.cherry_pick(branch if branch else ctx.nessie.get_default_branch(), condition, *hashes)
    click.echo()


@cli.command("contents")
@click.option(
    "-l",
    "--list",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="list tables",
    mutually_exclusive=["delete", "set"],
)
@click.option(
    "-d",
    "--delete",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="delete a table",
    mutually_exclusive=["list", "set"],
)
@click.option(
    "-s",
    "--set",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    help="modify a table",
    mutually_exclusive=["list", "delete"],
)
@click.option(
    "-c",
    "--condition",
    help="Conditional Hash. Only perform the action if branch currently points to condition.",
)
@click.option("-r", "--ref", help="branch to list from. If not supplied the default branch from config is used")
@click.option("-m", "--message", help="commit message")
@click.option(
    "-t",
    "--type",
    "entity_types",
    help="entity types to filter on, if no entity types are passed then all types are returned",
    multiple=True,
)
@click.option(
    "--query",
    "--query-expression",
    "query_expression",
    multiple=False,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["entity_type"],
    help="Allows advanced filtering using the Common Expression Language (CEL). "
    "An intro to CEL can be found at https://github.com/google/cel-spec/blob/master/doc/intro.md.\n"
    "Some examples with usable variables 'entry.namespace' (string) & 'entry.contentType' (string) are:\n"
    "entry.namespace.startsWith('a.b.c')\n"
    "entry.contentType in ['ICEBERG_TABLE','DELTA_LAKE_TABLE']\n"
    "entry.namespace.startsWith('some.name.space') && entry.contentType in ['ICEBERG_TABLE','DELTA_LAKE_TABLE']\n",
)
@click.option("--author", help="The author to use for the commit")
@click.argument("key", nargs=-1, required=False)
@pass_client
@error_handler
def contents(
    ctx: ContextObject,
    list: bool,
    delete: bool,
    set: bool,
    key: List[str],
    ref: str,
    message: str,
    condition: str,
    entity_types: List[str],
    author: str,
    query_expression: str,
) -> None:
    """Contents operations.

    KEY name of object to view, delete. If listing the key will limit by namespace what is included.
    """
    if list:
        keys = ctx.nessie.list_keys(
            ref if ref else ctx.nessie.get_default_branch(),
            query_expression=build_query_expression_for_contents_listing_flags(query_expression, entity_types),
        )
        results = EntrySchema().dumps(_format_keys_json(keys, *key), many=True) if ctx.json else _format_keys(keys, *key)
    elif delete:
        ctx.nessie.commit(ref, condition, _get_message(message), author, *_get_contents(ctx.nessie, ref, delete, *key))
        results = ""
    elif set:
        ctx.nessie.commit(ref, condition, _get_message(message), author, *_get_contents(ctx.nessie, ref, delete, *key))
        results = ""
    else:

        def content(*x: str) -> Generator[Contents, Any, None]:
            return ctx.nessie.get_values(ref if ref else ctx.nessie.get_default_branch(), *x)

        results = ContentsSchema().dumps(content(*key), many=True) if ctx.json else "\n".join((i.pretty_print() for i in content(*key)))
    click.echo(results)


def _get_contents(nessie: NessieClient, ref: str, delete: bool = False, *keys: str) -> List[Operation]:
    contents_altered: List[Operation] = list()
    for raw_key in keys:
        key = _format_key(raw_key)
        try:
            content = nessie.get_values(ref, key)
            content_json = ContentsSchema().dumps(content.__next__())
        except NessieNotFoundException:
            content_json = click.get_text_stream("stdin").read()

        MARKER = "# Everything below is ignored\n"
        message: Optional[str] = None
        if delete:
            message = "\n\n" + MARKER
        else:
            edit_message = (
                content_json
                + "\n\n"
                + MARKER
                + "Edit the content above to commit changes."
                + " Closing without change will result in a no-op."
                + " Removing the content results in a delete"
            )
            try:
                message = click.edit(edit_message)
            except click.ClickException:
                message = edit_message

        if message is not None:
            message_altered = message.split(MARKER, 1)[0].strip("\n")
            if message_altered:
                contents_altered.append(Put(_contents_key(raw_key), ContentsSchema().loads(message_altered)))
            else:
                contents_altered.append(Delete(_contents_key(raw_key)))
    return contents_altered


def _get_commit_message(*keys: str) -> Optional[str]:
    MARKER = "# Everything below is ignored\n"
    message = click.edit("\n\n" + MARKER + "\n".join(keys))
    return message.split(MARKER, 1)[0].rstrip("\n") if message is not None else None


def _get_message(message: str, *keys: str) -> Optional[str]:
    if message:
        return message
    return _get_commit_message(*keys)


def _format_keys_json(keys: Entries, *expected_keys: str) -> List[Entry]:
    results = list()
    for k in keys.entries:
        value = ['"{}"'.format(i) if "." in i else i for i in k.name.elements]
        if expected_keys and all(key for key in expected_keys if key not in value):
            continue
        results.append(k)
    return results


def _format_keys(keys: Entries, *key: str) -> str:
    results = defaultdict(list)
    result_str = ""
    for e in keys.entries:
        results[e.kind].append(e.name)
    for k in results.keys():
        result_str += k + ":\n"
        for v in results[k]:
            value = ['"{}"'.format(i) if "." in i else i for i in v.elements]
            if key and all(k for k in key if k not in value):
                continue
            result_str += "\t{}\n".format(".".join(value))
    return result_str


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



# -*- coding: utf-8 -*-
from typing import Optional

import click

from . import NessieClient
from .error import NessieNotFoundException
from .model import Branch
from .model import ReferenceSchema
from .model import Tag


def handle_branch_tag(
    nessie: NessieClient,
    list_references: bool,
    delete_reference: bool,
    branch: str,
    new_branch: str,
    is_branch: bool,
    json: bool,
    force: bool,
    verbose: bool,
    old_hash: Optional[str] = None,
) -> str:
    """Perform branch/tag actions."""
    if list_references or (not list_references and not delete_reference and not branch and not new_branch):
        return _handle_list(nessie, json, verbose, is_branch, branch)
    elif delete_reference:
        branch_object = nessie.get_reference(branch)
        getattr(nessie, "delete_{}".format("branch" if is_branch else "tag"))(branch, branch_object.hash_)
    elif branch and not new_branch:
        getattr(nessie, "create_{}".format("branch" if is_branch else "tag"))(branch)
    elif branch and new_branch and not force:

        try:
            getattr(nessie, "assign_{}".format("branch" if is_branch else "tag"))(branch, new_branch, old_hash)
        except NessieNotFoundException:
            create_on = nessie.get_reference(new_branch).hash_
            getattr(nessie, "create_{}".format("branch" if is_branch else "tag"))(branch, create_on)
    else:
        getattr(nessie, "assign_{}".format("branch" if is_branch else "tag"))(branch, new_branch, old_hash)
    return ""


def _handle_list(nessie: NessieClient, json: bool, verbose: bool, is_branch: bool, branch: str) -> str:
    results = nessie.list_references()
    kept_results = [ref for ref in results if isinstance(ref, (Branch if is_branch else Tag))]
    if json:
        return ReferenceSchema().dumps(kept_results, many=True)
    output = ""
    default_branch = nessie.get_default_branch()
    if branch:
        kept_results = [i for i in kept_results if i.name == branch]
    max_width = max((len(i.name) for i in kept_results), default=0)
    for x in kept_results:
        next_row = "{}{}{}{}{}\n".format(
            "*".ljust(2) if x.name == default_branch else "  ",
            x.name.ljust(max_width + 1),
            " " if verbose else "",
            x.hash_ if verbose else "",
            " comment" if verbose else "",
        )
        output += click.style(next_row, fg="yellow") if x.name == default_branch else next_row
    return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for `pynessie` package."""
import pytest

from pynessie import init
from pynessie.error import NessieConflictException
from pynessie.model import Branch
from pynessie.model import Entries


@pytest.mark.vcr
def test_client_interface_e2e() -> None:
    """Test client object against live server."""
    client = init()
    assert isinstance(client._base_url, str)
    assert client._base_url == "http://localhost:19120/api/v1"
    references = client.list_references()
    assert len(references) == 1
    assert references[0] == Branch("main", references[0].hash_)
    main_commit = references[0].hash_
    with pytest.raises(NessieConflictException):
        client.create_branch("main")
    created_reference = client.create_branch("test", main_commit)
    references = client.list_references()
    assert len(references) == 2
    assert next(i for i in references if i.name == "main") == Branch("main", main_commit)
    assert next(i for i in references if i.name == "test") == Branch("test", main_commit)
    reference = client.get_reference("test")
    assert created_reference == reference
    tables = client.list_keys(reference.name)
    assert isinstance(tables, Entries)
    assert len(tables.entries) == 0
    client.delete_branch("test", main_commit)
    references = client.list_references()
    assert len(references) == 1



# -*- coding: utf-8 -*-

"""The purpose of this module is to provide different functions that can produce CEL Expressions based on given filtering parameters."""

#  Copyright (C) 2020 Dremio
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import timezone
from typing import List, Optional

from dateutil import parser


def build_query_expression_for_commit_log_flags(
    query_expression: str, authors: List[str], committers: List[str], since: Optional[str], until: Optional[str]
) -> Optional[str]:
    """Producs a CEL expression to be used for filtering the commit log based on the given filtering parameters."""
    if query_expression:
        return query_expression
    expressions = []

    if authors:
        expressions.append(_expression_for_commit_log_by_author(authors))
    if committers:
        expressions.append(_expression_for_commit_log_by_committer(committers))
    if since:
        expressions.append(_expression_for_commit_log_by_commit_time_after(parse_to_iso8601(since)))
    if until:
        expressions.append(_expression_for_commit_log_by_commit_time_before(parse_to_iso8601(until)))

    if len(expressions) > 0:
        return _and_join(expressions)

    return None


def build_query_expression_for_contents_listing_flags(query_expression: str, types: List[str]) -> Optional[str]:
    """Producs a CEL expression to be used for filtering the content entries based on the given filtering parameters."""
    if query_expression:
        return query_expression

    if types:
        return _expression_for_entries_by_types(types)

    return None


def _expression_for_commit_log_by_author(authors: List[str]) -> str:
    return __generate_expression(authors, "commit.author=='{}'")


def _expression_for_commit_log_by_committer(committers: List[str]) -> str:
    return __generate_expression(committers, "commit.committer=='{}'")


def _expression_for_commit_log_by_commit_time_after(commit_time_after: str) -> str:
    return "timestamp(commit.commitTime) > timestamp('{}')".format(commit_time_after)


def _expression_for_commit_log_by_commit_time_before(commit_time_before: str) -> str:
    return "timestamp(commit.commitTime) < timestamp('{}')".format(commit_time_before)


def _expression_for_entries_by_types(types: List[str]) -> str:
    return "entry.contentType in [{}]".format(",".join(["'" + t + "'" for t in types]))


def _and_join(expressions: List[str]) -> str:
    result = " && ".join(expressions)
    if len(expressions) > 1:
        result = "(" + result + ")"
    return result


def _or_join(expressions: List[str]) -> str:
    result = " || ".join(expressions)
    if len(expressions) > 1:
        result = "(" + result + ")"
    return result


def __generate_expression(items: List[str], expression_template: str) -> str:
    expressions = []
    for item in items:
        expressions.append(expression_template.format(item))
    return _or_join(expressions)


def parse_to_iso8601(date_str: str) -> str:
    """Parses a given Date string to an ISO 8601 compliant string in UTC.

    :param date_str A string representation of a date
    :return An ISO 8601 string in the form YYYY-MM-DD HH:MM:SS.mmmmmm+HH:MM
    """
    try:
        return parser.parse(date_str).astimezone(tz=timezone.utc).isoformat(sep="T")
    except Exception as e:
        raise ValueError("Unable to parse '{}' to an ISO 8601-compliant string. Reason: {}".format(date_str, e), e)



# -*- coding: utf-8 -*-
"""Main module."""
import re
from typing import Any
from typing import cast
from typing import Generator
from typing import List
from typing import Optional

import confuse

from ._endpoints import all_references
from ._endpoints import assign_branch
from ._endpoints import assign_tag
from ._endpoints import cherry_pick
from ._endpoints import commit
from ._endpoints import create_reference
from ._endpoints import delete_branch
from ._endpoints import delete_tag
from ._endpoints import get_default_branch
from ._endpoints import get_reference
from ._endpoints import get_table
from ._endpoints import list_logs
from ._endpoints import list_tables
from ._endpoints import merge
from .model import Branch
from .model import CommitMeta
from .model import Contents
from .model import ContentsKey
from .model import ContentsSchema
from .model import Entries
from .model import EntriesSchema
from .model import LogResponse
from .model import LogResponseSchema
from .model import Merge
from .model import MergeSchema
from .model import MultiContents
from .model import MultiContentsSchema
from .model import Operation
from .model import Reference
from .model import ReferenceSchema
from .model import Tag
from .model import Transplant
from .model import TransplantSchema

_dot_regex = re.compile('\\.(?=([^"]*"[^"]*")*[^"]*$)')


class NessieClient(object):
    """Base Nessie Client."""

    def __init__(self: "NessieClient", config: confuse.Configuration) -> None:
        """Create a Nessie Client from known config."""
        url = config["endpoint"].get()
        repo = config["repo"].get() if "repo" in config else None
        if repo:
            sep = "" if url[-1:] == "/" else "/"
            url = f"{url}{sep}{repo}"
        self._base_url = url
        self._ssl_verify = config["verify"].get(bool)
        self._commit_id: str = cast(str, None)

        try:
            self._base_branch = config["default_branch"].get()
        except confuse.exceptions.NotFoundError:
            self._base_branch = None

    def list_references(self: "NessieClient") -> List[Reference]:
        """Fetch all known references.

        :return: list of Nessie References
        """
        references = all_references(self._base_url, self._ssl_verify)
        return [ReferenceSchema().load(ref) for ref in references]

    def get_reference(self: "NessieClient", name: Optional[str]) -> Reference:
        """Fetch a ref.

        :param name: name of ref to fetch
        :return: Nessie reference
        """
        ref_obj = get_reference(self._base_url, name, self._ssl_verify) if name else get_default_branch(self._base_url, self._ssl_verify)
        ref = ReferenceSchema().load(ref_obj)
        return ref

    def create_branch(self: "NessieClient", branch: str, ref: str = None) -> Branch:
        """Create a branch.

        :param branch: name of new branch
        :param ref: ref to fork from
        :return: Nessie branch object
        """
        ref_json = ReferenceSchema().dump(Branch(branch, ref))
        ref_obj = create_reference(self._base_url, ref_json, self._ssl_verify)
        return cast(Branch, ReferenceSchema().load(ref_obj))

    def delete_branch(self: "NessieClient", branch: str, hash_: str) -> None:
        """Delete a branch.

        :param branch: name of branch to delete
        :param hash_: hash of the branch
        """
        delete_branch(self._base_url, branch, hash_, self._ssl_verify)

    def create_tag(self: "NessieClient", tag: str, ref: str = None) -> Tag:
        """Create a tag.

        :param tag: name of new tag
        :param ref: ref to fork from
        :return: Nessie tag object
        """
        ref_json = ReferenceSchema().dump(Tag(tag, ref))
        ref_obj = create_reference(self._base_url, ref_json, self._ssl_verify)
        return cast(Tag, ReferenceSchema().load(ref_obj))

    def delete_tag(self: "NessieClient", tag: str, hash_: str) -> None:
        """Delete a tag.

        :param tag: name of tag to delete
        :param hash_: hash of the branch
        """
        delete_tag(self._base_url, tag, hash_, self._ssl_verify)

    def list_keys(
        self: "NessieClient",
        ref: str,
        max_result_hint: Optional[int] = None,
        page_token: Optional[str] = None,
        query_expression: Optional[str] = None,
    ) -> Entries:
        """Fetch a list of all tables from a known branch.

        :param ref: name of branch
        :param entity_types: list of types to filter keys on
        :param query_expression: A CEL expression that allows advanced filtering capabilities
        :return: list of Nessie table names
        """
        return EntriesSchema().load(list_tables(self._base_url, ref, max_result_hint, page_token, query_expression, self._ssl_verify))

    def get_values(self: "NessieClient", ref: str, *tables: str) -> Generator[Contents, Any, None]:
        """Fetch a table from a known ref.

        :param ref: name of ref
        :param tables: tables to fetch
        :return: Nessie Table
        """
        return (ContentsSchema().load(get_table(self._base_url, ref, _format_key(i), self._ssl_verify)) for i in tables)

    def commit(
        self: "NessieClient", branch: str, old_hash: str, reason: Optional[str] = None, author: Optional[str] = None, *ops: Operation
    ) -> Branch:
        """Modify a set of Nessie tables."""
        meta = CommitMeta(message=reason if reason else "")
        if author:
            meta.author = author
        ref_obj = commit(self._base_url, branch, MultiContentsSchema().dumps(MultiContents(meta, list(ops))), old_hash)
        return cast(Branch, ReferenceSchema().load(ref_obj))

    def assign_branch(self: "NessieClient", branch: str, to_ref: str, old_hash: Optional[str] = None) -> None:
        """Assign a hash to a branch."""
        if not old_hash:
            old_hash = self.get_reference(branch).hash_
        assert old_hash is not None
        branch_json = ReferenceSchema().dumps(Branch(branch, self.get_reference(to_ref).hash_))
        assign_branch(self._base_url, branch, branch_json, old_hash, self._ssl_verify)

    def assign_tag(self: "NessieClient", tag: str, to_ref: str, old_hash: Optional[str] = None) -> None:
        """Assign a hash to a tag."""
        if not old_hash:
            old_hash = self.get_reference(tag).hash_
        assert old_hash is not None
        tag_json = ReferenceSchema().dumps(Tag(tag, self.get_reference(to_ref).hash_))
        assign_tag(self._base_url, tag, tag_json, old_hash, self._ssl_verify)

    def merge(self: "NessieClient", from_branch: str, onto_branch: str, old_hash: Optional[str] = None) -> None:
        """Merge a branch into another branch."""
        if not old_hash:
            old_hash = self.get_reference(onto_branch).hash_
        assert old_hash is not None
        from_hash = self.get_reference(from_branch).hash_
        assert from_hash is not None
        merge_json = MergeSchema().dump(Merge(from_hash))
        merge(self._base_url, onto_branch, merge_json, old_hash, self._ssl_verify)

    def cherry_pick(self: "NessieClient", branch: str, old_hash: Optional[str] = None, *hashes: str) -> None:
        """Cherry pick a list of hashes to a branch."""
        if not old_hash:
            old_hash = self.get_reference(branch).hash_
        assert old_hash is not None
        transplant_json = TransplantSchema().dump(Transplant(list(hashes)))
        cherry_pick(self._base_url, branch, transplant_json, old_hash, self._ssl_verify)

    def get_log(self: "NessieClient", start_ref: str, **filtering_args: Any) -> Generator[CommitMeta, Any, None]:
        """Fetch all logs starting at start_ref.

        start_ref can be any ref.

        Note:
            this will load the log into local memory and filter at the client. Currently there are no
            primitives in the REST api to limit logs or perform paging. TODO
        """
        page_token = filtering_args.get("pageToken", None)

        def fetch_logs(token: Optional[str] = page_token) -> LogResponse:
            if token:
                filtering_args["pageToken"] = token

            fetched_logs = list_logs(base_url=self._base_url, ref=start_ref, ssl_verify=self._ssl_verify, **filtering_args)
            log_schema = LogResponseSchema().load(fetched_logs)
            return log_schema

        log_schema = fetch_logs()

        def generator(log_schema: LogResponse) -> Generator[CommitMeta, Any, None]:
            while True:
                for i in log_schema.operations:
                    yield i
                if log_schema.has_more:
                    log_schema = fetch_logs(token=log_schema.token)
                else:
                    break

        return generator(log_schema)

    def get_default_branch(self: "NessieClient") -> str:
        """Fetch default branch either from config if specified or from the server."""
        return self._base_branch if self._base_branch else self.get_reference(None).name


def _format_key(raw_key: str) -> str:
    elements = _dot_regex.split(raw_key)
    return ".".join(i.replace(".", "\0") for i in elements if i)


def _contents_key(raw_key: str) -> ContentsKey:
    elements = _dot_regex.split(raw_key)
    return ContentsKey([i for i in elements if i])
